;; attribute.el
;; Set character attribute for Emacs under X window environment.
;; Copyright (C) 1989 Free Software Foundation, Inc.
;; Written by K. HANDA, Electrotechnical Lab., JAPAN, 1989.

;; This program is distributed under the terms of the GNU
;; EMACS GENERAL PUBLIC LICENSE which is distributed along
;; with GNU Emacs by the Free Software Foundation.

;; This program is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY; without even the
;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;; PARTICULAR PURPOSE.  See the GNU EMACS GENERAL PUBLIC
;; LICENSE for more details.

;; You should have received a copy of the GNU EMACS GENERAL
;; PUBLIC LICENSE along with this file; see the file
;; COPYING.  If not, write to the Free Software Foundation,
;; 675 Mass Ave, Cambridge, MA 02139, USA.

;; Change Log
;; 89.11.28 Created by K.Handa
;; 89.12.10 Update screen after attributes are changed.  by K.Handa

(provide 'attribute)

(defun attribute-on-region (attr &optional from to)
  "Set an charater attribute ATTR within the region on.\n\
Currently supported attributes are 'underline' and 'inverse'.\n\
Called from program, takes two optional arguments FROM and TO.\n\
If FROM or TO is NIL, (point-min) or (point-max) is used instead.\n\
Information of attribute is stored in buffer local variables\n\
underlined-region and inversed-region, but manipulating these variabels\n\
directly may cause serious problems."
  (interactive
   (list
    (attribute-symbol-read)
    (mark)
    (point)))
  (attribute-on-off-region attr from to 'on))

(defun attribute-off-region (attr &optional from to)
  "Set an charater attribute ATTR within the region off.\n\
See the function attribute-on-region for the detail."
  (interactive
   (list
    (attribute-symbol-read)
    (mark)
    (point)))
  (attribute-on-off-region attr from to nil))

(defun attribute-symbol-read ()
  (intern (completing-read
	   "Attribute: " '(("inverse") ("underline")) nil t)))

(defun attribute-get-attribute (&optional pos)
  "Return a list of attribute ranges at POS.\n\
An attribute range is a list of attribute type, the starting point,\n\
and the ending point.\n\
POS is optional argument and the default is the current point."
  (if (not pos) (setq pos (point)))
  (let (return attr)
    (if (and underlined-region
	     (setq attr (attribute-starting-point underlined-region pos)))
	(setq return (cons (list 'underline (car attr) (car (cdr attr)))
			   return)))
    (if (and inversed-region
	     (setq attr (attribute-starting-point inversed-region pos)))
	(setq return (cons (list 'inverse (car attr) (car (cdr attr)))
			   return)))
    return))

(defun attribute-starting-point (attr pos)
  (while (and (setq attr (cdr attr))
	      (>= pos (car (cdr attr))))
    (setq attr (cdr attr)))
  (if attr
      (if (>= pos (car attr)) attr)))

(defun attribute-on-off-region (attr &optional from to on)
  (let ((min (point-min)) (max (point-max)))
    (if (or (not from) (< from min)) (setq from min)
      (if (> from max) (setq from max)))
    (if (or (not to) (> to max)) (setq to max)
      (if (< to min) (setq to min))))
  (if (> from to) (let ((temp from)) (setq from to) (setq to temp)))
  (attribute-record-region-modified from to)
  (let* ((attribute (cond ((eq attr 'underline) 'underlined-region)
			  ((eq attr 'inverse) 'inversed-region)
			  ;;; ((eq attr 'highlight) 'highlighted-region)
			  (t nil))))
    (if attribute
	(progn
	  (if (not (symbol-value attribute))
	      (set attribute (list 1)))
	  (attribute-modify-attribute attribute from to (if on 1 0)))
      (error "Invalid ATTR(%s)" attr))))

(defun attribute-modify-attribute (attr from to set)
  (let* ((attr-value (symbol-value attr))
	 (len (car attr-value))
	 (i1 1) i2)
;    (let (m1 m2)			; at first, erase invalid block
;      (while (< i1 len)
;	(setq m1 (nth i1 attr-value) m2 (nth (1+ i1) attr-value))
;	(if (or (not m1) (not m2) (= m1 m2))
;	    (progn
;	      (set-marker m1 nil)
;	      (set-marker m2 nil)
;	      (rplacd (nthcdr (1- i1) attr-value)
;		      (nthcdr (+ i1 2) attr-value))
;	      (rplaca attr-value (- len 2)))
;	  (setq i1 (+ i1 2)))))
    (setq i1 1)
    (while (and (< i1 len) (> from (nth i1 attr-value)))
      (setq i1 (1+ i1)))
    (setq i2 i1)
    (while (and (< i2 len) (>= to (nth i2 attr-value)))
      (setq i2 (1+ i2)))
    (if (= i1 i2)			; if within the same area
	(if (= (% i1 2) set)		;   if within the same attribute area
	    (let ((m1 (make-marker)) (m2 (make-marker)))
	      (set-marker m1 from)
	      (set-marker m2 to)
	      ; (set-marker-type m2 'point-type) ; for Nemacs
	      (rplacd (nthcdr (1- i1) attr-value)
		      (cons m1 (cons m2 (nthcdr i1 attr-value))))))
					; else laid across block border
      (if (= (% i1 2) set)		;   if 'from' is out of attribute area
	  (move-marker (nth i1 attr-value) from)
	(setq i1 (1- i1)))
      (if (= (% i2 2) set)		;   if 'to' is out of attribute area
	  (progn
	    (setq i2 (1- i2))
	    (move-marker (nth i2 attr-value) to)))
      (if (< i1 (1- i2))		;   if the list should be reduced
	  (let ((i i1))
	    (while (< (setq i (1+ i)) i2)
	      (set-marker (nth i attr-value) nil))
	    (rplacd (nthcdr i1 attr-value) (nthcdr i2 attr-value)))))
    (rplaca attr-value (length attr-value))
    (if (= (car attr-value) 1) (set attr nil))))

;; There should be a better way! :-(
(defun attribute-record-region-modified (from to)
  (subst-char-in-region from to 0 0 'noundo))
